"""
Unit or integration tests for gpustat
"""
# flake8: ignore=E501

from __future__ import print_function
from __future__ import absolute_import

import sys
import os
import shlex
from collections import namedtuple
from io import StringIO    # Python 3 only

import psutil
import pytest
from mockito import when, mock, unstub

import gpustat
from gpustat.nvml import pynvml


MB = 1024 * 1024


def remove_ansi_codes(s):
    import re
    s = re.compile(r'\x1b[^m]*m').sub('', s)
    s = re.compile(r'\x0f').sub('', s)
    return s

# -----------------------------------------------------------------------------


def _configure_mock(N=pynvml, scenario_nonexistent_pid=False,
                    nosuchprocess_exception_type=psutil.NoSuchProcess,
                    ):
    """Define mock behaviour for pynvml and psutil.{Process,virtual_memory}."""

    # without following patch, unhashable NVMLError makes unit test crash
    N.NVMLError.__hash__ = lambda _: 0
    assert issubclass(N.NVMLError, BaseException)

    when(N).nvmlInit().thenReturn()
    when(N).nvmlShutdown().thenReturn()
    when(N).nvmlSystemGetDriverVersion().thenReturn('415.27.mock')

    NUM_GPUS = 3
    mock_handles = ['mock-handle-%d' % i for i in range(3)]
    when(N).nvmlDeviceGetCount().thenReturn(NUM_GPUS)

    def _return_or_raise(v):
        """Return a callable for thenAnswer() to let exceptions re-raised."""
        def _callable(*args, **kwargs):
            if isinstance(v, Exception):
                raise v
            return v
        return _callable

    for i in range(NUM_GPUS):
        handle = mock_handles[i]
        when(N).nvmlDeviceGetHandleByIndex(i)\
            .thenReturn(handle)
        when(N).nvmlDeviceGetIndex(handle)\
            .thenReturn(i)
        when(N).nvmlDeviceGetName(handle)\
            .thenReturn({
                0: 'GeForce GTX TITAN 0',
                1: 'GeForce GTX TITAN 1',
                2: 'GeForce RTX 2',
            }[i].encode())
        when(N).nvmlDeviceGetUUID(handle)\
            .thenReturn({
                0: b'GPU-10fb0fbd-2696-43f3-467f-d280d906a107',
                1: b'GPU-d1df4664-bb44-189c-7ad0-ab86c8cb30e2',
                2: b'GPU-50205d95-57b6-f541-2bcb-86c09afed564',
            }[i])

        when(N).nvmlDeviceGetTemperature(handle, N.NVML_TEMPERATURE_GPU)\
            .thenReturn([80, 36, 71][i])
        when(N).nvmlDeviceGetFanSpeed(handle)\
            .thenReturn([16, 53, 100][i])
        when(N).nvmlDeviceGetPowerUsage(handle)\
            .thenAnswer(_return_or_raise({
                0: 125000, 1: N.NVMLError_NotSupported(), 2: 250000
            }[i]))
        when(N).nvmlDeviceGetEnforcedPowerLimit(handle)\
            .thenAnswer(_return_or_raise({
                0: 250000, 1: 250000, 2: N.NVMLError_NotSupported()
            }[i]))

        mock_memory_t = namedtuple("Memory_t", ['total', 'used'])
        when(N).nvmlDeviceGetMemoryInfo(handle)\
            .thenAnswer(_return_or_raise({
                0: mock_memory_t(total=12883853312, used=8000*MB),
                1: mock_memory_t(total=12781551616, used=9000*MB),
                2: mock_memory_t(total=12781551616, used=0),
            }[i]))

        mock_utilization_t = namedtuple("Utilization_t", ['gpu', 'memory'])
        when(N).nvmlDeviceGetUtilizationRates(handle)\
            .thenAnswer(_return_or_raise({
                0: mock_utilization_t(gpu=76, memory=0),
                1: mock_utilization_t(gpu=0, memory=0),
                2: N.NVMLError_NotSupported(),  # Not Supported
            }[i]))

        when(N).nvmlDeviceGetEncoderUtilization(handle)\
            .thenAnswer(_return_or_raise({
                0: [88, 167000],  # [value, sample_rate]
                1: [0, 167000],   # [value, sample_rate]
                2: N.NVMLError_NotSupported(),  # Not Supported
            }[i]))
        when(N).nvmlDeviceGetDecoderUtilization(handle)\
            .thenAnswer(_return_or_raise({
                0: [67, 167000],  # [value, sample_rate]
                1: [0, 167000],   # [value, sample_rate]
                2: N.NVMLError_NotSupported(),  # Not Supported
            }[i]))

        # running process information: a bit annoying...
        mock_process_t = namedtuple("Process_t", ['pid', 'usedGpuMemory'])

        if scenario_nonexistent_pid:
            mock_processes_gpu2_erratic = [
                mock_process_t(99999, 9999*MB),
                mock_process_t(99995, 9995*MB),   # see issue #95
            ]
        else:
            mock_processes_gpu2_erratic = N.NVMLError_NotSupported()
        when(N).nvmlDeviceGetComputeRunningProcesses(handle)\
            .thenAnswer(_return_or_raise({
                0: [mock_process_t(48448, 4000*MB), mock_process_t(153223, 4000*MB)],
                1: [mock_process_t(192453, 3000*MB), mock_process_t(194826, 6000*MB)],
                2: mock_processes_gpu2_erratic,   # Not Supported or non-existent
            }[i]))

        when(N).nvmlDeviceGetGraphicsRunningProcesses(handle)\
            .thenAnswer(_return_or_raise({
                0: [mock_process_t(48448, 4000*MB)],
                1: [],
                2: N.NVMLError_NotSupported(),
            }[i]))

    # for psutil
    mock_pid_map = {   # mock/stub information for psutil...
        48448:  ('user1', 'python', 85.25, 3.1415),
        154213: ('user1', 'caffe', 16.89, 100.00),
        38310:  ('user3', 'python', 26.23, 99.9653),
        153223: ('user2', 'python', 15.25, 0.0000),
        194826: ('user3', 'caffe', 0.0, 12.5236),
        192453: ('user1', 'torch', 123.2, 0.7312),
    }
    assert 99999 not in mock_pid_map, 'scenario_nonexistent_pid'
    assert 99995 not in mock_pid_map, 'scenario_nonexistent_pid (#95)'

    def _MockedProcess(pid):
        if pid not in mock_pid_map:
            if pid == 99995:
                # simulate a bug reported in #95
                raise FileNotFoundError("/proc/99995/stat")
            else:
                # for a process that does not exist, NoSuchProcess is the
                # type of exceptions supposed to be raised by psutil
                raise psutil.NoSuchProcess(pid=pid)
        username, cmdline, cpuutil, memutil = mock_pid_map[pid]
        p = mock(strict=True)
        p.username = lambda: username
        p.cmdline = lambda: [cmdline]
        p.cpu_percent = lambda: cpuutil
        p.memory_percent = lambda: memutil
        p.pid = pid
        return p

    when(psutil).Process(...)\
        .thenAnswer(_MockedProcess)
    when(psutil).virtual_memory()\
        .thenReturn(mock_memory_t(total=8589934592, used=0))


MOCK_EXPECTED_OUTPUT_DEFAULT = os.linesep.join("""\
[0] GeForce GTX TITAN 0 | 80°C,  76 % |  8000 / 12287 MB | user1(4000M) user2(4000M)
[1] GeForce GTX TITAN 1 | 36°C,   0 % |  9000 / 12189 MB | user1(3000M) user3(6000M)
[2] GeForce RTX 2       | 71°C,  ?? % |     0 / 12189 MB | (Not Supported)
""".splitlines())  # noqa: E501

MOCK_EXPECTED_OUTPUT_FULL = os.linesep.join("""\
[0] GeForce GTX TITAN 0 | 80°C,  16 %,  76 % (E:  88 %  D:  67 %),  125 / 250 W |  8000 / 12287 MB | user1:python/48448(4000M) user2:python/153223(4000M)
[1] GeForce GTX TITAN 1 | 36°C,  53 %,   0 % (E:   0 %  D:   0 %),   ?? / 250 W |  9000 / 12189 MB | user1:torch/192453(3000M) user3:caffe/194826(6000M)
[2] GeForce RTX 2       | 71°C, 100 %,  ?? % (E:  ?? %  D:  ?? %),  250 /  ?? W |     0 / 12189 MB | (Not Supported)
""".splitlines())  # noqa: E501

MOCK_EXPECTED_OUTPUT_FULL_PROCESS = os.linesep.join("""\
[0] GeForce GTX TITAN 0 | 80°C,  16 %,  76 % (E:  88 %  D:  67 %),  125 / 250 W |  8000 / 12287 MB | user1:python/48448(4000M) user2:python/153223(4000M)
 ├─  48448 (  85%,  257MB): python
 └─ 153223 (  15%,     0B): python
[1] GeForce GTX TITAN 1 | 36°C,  53 %,   0 % (E:   0 %  D:   0 %),   ?? / 250 W |  9000 / 12189 MB | user1:torch/192453(3000M) user3:caffe/194826(6000M)
 ├─ 192453 ( 123%,   59MB): torch
 └─ 194826 (   0%, 1025MB): caffe
[2] GeForce RTX 2       | 71°C, 100 %,  ?? % (E:  ?? %  D:  ?? %),  250 /  ?? W |     0 / 12189 MB | (Not Supported)
""".splitlines())  # noqa: E501


# -----------------------------------------------------------------------------

@pytest.fixture
def scenario_basic():
    _configure_mock()

@pytest.fixture
def scenario_nonexistent_pid():
    _configure_mock(scenario_nonexistent_pid=True)


class TestGPUStat(object):
    """A pytest class suite for gpustat."""

    def setup_method(self):
        print("")
        self.maxDiff = 4096

    def teardown_method(self):
        unstub()

    @staticmethod
    def capture_output(*args):
        f = StringIO()
        import contextlib

        with contextlib.redirect_stdout(f):  # requires python 3.4+
            try:
                gpustat.main(*args)
            except SystemExit as e:
                if e.code != 0:
                    raise AssertionError(
                        "Argparse failed (see above error message)")
        return f.getvalue()

    # -----------------------------------------------------------------------

    def test_new_query_mocked(self, scenario_basic):
        """
        A basic functionality test, in a case where everything is just normal.
        """

        gpustats = gpustat.new_query()
        fp = StringIO()
        gpustats.print_formatted(
            fp=fp, no_color=False, show_user=True,
            show_cmd=True, show_full_cmd=True, show_pid=True,
            show_fan_speed=True, show_codec="enc,dec", show_power=True,
        )

        result = fp.getvalue()
        print(result)

        unescaped = remove_ansi_codes(result)
        # remove first line (header)
        unescaped = os.linesep.join(unescaped.splitlines()[1:])

        assert unescaped == MOCK_EXPECTED_OUTPUT_FULL_PROCESS

    def test_new_query_mocked_nonexistent_pid(self, scenario_nonexistent_pid):
        """
        Test a case where nvidia query returns non-existent pids (see #16, #18)
        for GPU index 2.
        """
        fp = StringIO()

        gpustats = gpustat.new_query()
        gpustats.print_formatted(fp=fp)

        ret = fp.getvalue()
        print(ret)

        # gpu 2: should ignore process id
        line = remove_ansi_codes(ret).split('\n')[3]
        assert '[2] GeForce RTX 2' in line, str(line)
        assert '99999' not in line
        assert '(Not Supported)' not in line

    def test_attributes_and_items(self, scenario_basic):
        """Test whether each property of `GPUStat` instance is well-defined."""

        g = gpustat.new_query()[1]  # includes N/A
        print("(keys) : %s" % str(g.keys()))
        print(g)

        assert g['name'] == g.entry['name']
        assert g['uuid'] == g.uuid

        with pytest.raises(KeyError):
            g['unknown_key']

        print("uuid : %s" % g.uuid)
        print("name : %s" % g.name)
        print("memory : used %d total %d avail %d" % (
            g.memory_used, g.memory_total, g.memory_available))
        print("temperature : %d" % (g.temperature))
        print("utilization : %s" % (g.utilization))
        print("utilization_enc : %s" % (g.utilization_enc))
        print("utilization_dec : %s" % (g.utilization_dec))

    def test_main(self, scenario_basic):
        """Test whether gpustat.main() works well.
        The behavior is mocked exactly as in test_new_query_mocked().
        """
        sys.argv = ['gpustat']
        gpustat.main()

    def test_args_commandline(self, scenario_basic):
        """Tests the end gpustat CLI."""
        capture_output = self.capture_output

        def _remove_ansi_codes_and_header_line(s):
            unescaped = remove_ansi_codes(s)
            # remove first line (header)
            unescaped = os.linesep.join(unescaped.splitlines()[1:])
            return unescaped

        s = capture_output('gpustat', )
        assert _remove_ansi_codes_and_header_line(s) == MOCK_EXPECTED_OUTPUT_DEFAULT

        s = capture_output('gpustat', '--version')
        assert s.startswith('gpustat ')
        print(s)

        s = capture_output('gpustat', '--no-header')
        assert "[0]" in s.splitlines()[0]

        s = capture_output('gpustat', '-a')  # --show-all
        assert _remove_ansi_codes_and_header_line(s) == MOCK_EXPECTED_OUTPUT_FULL

        s = capture_output('gpustat', '--color')
        assert '\x0f' not in s, "Extra \\x0f found (see issue #32)"
        assert _remove_ansi_codes_and_header_line(s) == MOCK_EXPECTED_OUTPUT_DEFAULT

        s = capture_output('gpustat', '--no-color')
        unescaped = remove_ansi_codes(s)
        assert s == unescaped   # should have no ansi code
        assert _remove_ansi_codes_and_header_line(s) == MOCK_EXPECTED_OUTPUT_DEFAULT

    def test_args_commandline_width(self, scenario_basic):
        capture_output = self.capture_output

        # see MOCK_EXPECTED_OUTPUT_DEFAULT
        assert len("GeForce GTX TITAN 0") == 19

        s = capture_output('gpustat', '--gpuname-width', '25')
        print("- Should have width=25")
        print(s)
        assert 'GeForce GTX TITAN 0       |' in remove_ansi_codes(s)
        assert 'GeForce RTX 2             |' in remove_ansi_codes(s)
        #                         ^012345
        #                        19

        # See #47 (since v1.0)
        print("- Should have width=10 (with truncation)")
        s = capture_output('gpustat', '--gpuname-width', '10')
        print(s)
        assert '…X TITAN 0 |' in remove_ansi_codes(s)
        assert '…rce RTX 2 |' in remove_ansi_codes(s)
        #       1234567890

        print("- Should have width=1 (too short)")
        s = capture_output('gpustat', '--gpuname-width', '1')
        print(s)
        assert '… |' in remove_ansi_codes(s)

        print("- Should have width=0: no name displayed.")
        s = capture_output('gpustat', '--gpuname-width', '0')
        print(s)
        assert '[0]  80°C' in remove_ansi_codes(s)

        print("- Invalid inputs")
        with pytest.raises(AssertionError, match="Argparse failed"):
            s = capture_output('gpustat', '--gpuname-width', '-1')
        with pytest.raises(AssertionError, match="Argparse failed"):
            s = capture_output('gpustat', '--gpuname-width', 'None')

    def test_args_commandline_showoptions(self, scenario_basic):
        """Tests gpustat CLI with a variety of --show-xxx options. """

        capture_output = self.capture_output
        print('')

        TEST_OPTS = []
        TEST_OPTS += ['-a', '-c', '-u', '-p', '-e', '-P', '-f']
        TEST_OPTS += [('-e', ''), ('-P', '')]
        TEST_OPTS += [('-e', 'enc,dec'), '-Plimit,draw']
        TEST_OPTS += ['-cup', '-cpu', '-cufP']  # 'cpuePf'

        for opt in TEST_OPTS:
            if isinstance(opt, str):
                opt = [opt]

            print('\x1b[30m\x1b[43m',  # black_on_yellow
                  '$ gpustat ' + ' '.join(shlex.quote(o) for o in opt),
                  '\x1b(B\x1b[m', sep='')
            s = capture_output('gpustat', *opt)

            # TODO: Validate output without hardcoding expected outputs
            print(s)

        # Finally, unknown args
        with pytest.raises(AssertionError):
            capture_output('gpustat', '--unrecognized-args-in-test')

    @pytest.mark.skipif(sys.platform == 'win32', reason="Do not run on Windows")
    def test_no_TERM(self, scenario_basic, monkeypatch):
        """--color should work well even when executed without TERM,
        e.g. ssh localhost gpustat --color"""
        monkeypatch.setenv("TERM", "")

        s = self.capture_output('gpustat', '--color', '--no-header').rstrip()
        print(s)
        assert remove_ansi_codes(s) == MOCK_EXPECTED_OUTPUT_DEFAULT, \
            "wrong gpustat output"

        assert '\x1b[36m' in s, "should contain cyan color code"
        assert '\x0f' not in s, "Extra \\x0f found (see issue #32)"

    def test_json_mocked(self, scenario_basic):
        gpustats = gpustat.new_query()

        fp = StringIO()
        gpustats.print_json(fp=fp)

        import json
        j = json.loads(fp.getvalue())

        from pprint import pprint
        pprint(j)

        assert j['driver_version'] == '415.27.mock'
        assert j['hostname']
        assert j['gpus']


if __name__ == '__main__':
    pytest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Implementation of gpustat

@author Jongwook Choi
@url https://github.com/wookayin/gpustat
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import json
import locale
import os.path
import platform
import sys
import time
from datetime import datetime

from six.moves import cStringIO as StringIO
import psutil
from blessed import Terminal

from gpustat.nvml import pynvml as N
import gpustat.util as util


NOT_SUPPORTED = 'Not Supported'
MB = 1024 * 1024

DEFAULT_GPUNAME_WIDTH = 16

IS_WINDOWS = 'windows' in platform.platform().lower()


class GPUStat(object):

    def __init__(self, entry):
        if not isinstance(entry, dict):
            raise TypeError(
                'entry should be a dict, {} given'.format(type(entry))
            )
        self.entry = entry

    def __repr__(self):
        return self.print_to(StringIO()).getvalue()

    def keys(self):
        return self.entry.keys()

    def __getitem__(self, key):
        return self.entry[key]

    @property
    def index(self):
        """
        Returns the index of GPU (as in nvidia-smi).
        """
        return self.entry['index']

    @property
    def uuid(self):
        """
        Returns the uuid returned by nvidia-smi,
        e.g. GPU-12345678-abcd-abcd-uuid-123456abcdef
        """
        return self.entry['uuid']

    @property
    def name(self):
        """
        Returns the name of GPU card (e.g. Geforce Titan X)
        """
        return self.entry['name']

    @property
    def memory_total(self):
        """
        Returns the total memory (in MB) as an integer.
        """
        return int(self.entry['memory.total'])

    @property
    def memory_used(self):
        """
        Returns the occupied memory (in MB) as an integer.
        """
        return int(self.entry['memory.used'])

    @property
    def memory_free(self):
        """
        Returns the free (available) memory (in MB) as an integer.
        """
        v = self.memory_total - self.memory_used
        return max(v, 0)

    @property
    def memory_available(self):
        """
        Returns the available memory (in MB) as an integer.
        Alias of memory_free.
        """
        return self.memory_free

    @property
    def temperature(self):
        """
        Returns the temperature (in celcius) of GPU as an integer,
        or None if the information is not available.
        """
        v = self.entry['temperature.gpu']
        return int(v) if v is not None else None

    @property
    def fan_speed(self):
        """
        Returns the fan speed percentage (0-100) of maximum intended speed
        as an integer, or None if the information is not available.
        """
        v = self.entry['fan.speed']
        return int(v) if v is not None else None

    @property
    def utilization(self):
        """
        Returns the GPU utilization (in percentile),
        or None if the information is not available.
        """
        v = self.entry['utilization.gpu']
        return int(v) if v is not None else None

    @property
    def utilization_enc(self):
        """
        Returns the GPU encoder utilization (in percentile),
        or None if the information is not available.
        """
        v = self.entry['utilization.enc']
        return int(v) if v is not None else None

    @property
    def utilization_dec(self):
        """
        Returns the GPU decoder utilization (in percentile),
        or None if the information is not available.
        """
        v = self.entry['utilization.dec']
        return int(v) if v is not None else None

    @property
    def power_draw(self):
        """
        Returns the GPU power usage in Watts,
        or None if the information is not available.
        """
        v = self.entry['power.draw']
        return int(v) if v is not None else None

    @property
    def power_limit(self):
        """
        Returns the (enforced) GPU power limit in Watts,
        or None if the information is not available.
        """
        v = self.entry['enforced.power.limit']
        return int(v) if v is not None else None

    @property
    def processes(self):
        """
        Get the list of running processes on the GPU.
        """
        return self.entry['processes']

    def print_to(self, fp,
                 with_colors=True,    # deprecated arg
                 show_cmd=False,
                 show_full_cmd=False,
                 show_user=False,
                 show_pid=False,
                 show_fan_speed=None,
                 show_codec="",
                 show_power=None,
                 gpuname_width=None,
                 term=None,
                 ):
        if term is None:
            term = Terminal(stream=sys.stdout)

        # color settings
        colors = {}

        def _conditional(cond_fn, true_value, false_value,
                         error_value=term.bold_black):
            try:
                return cond_fn() and true_value or false_value
            except Exception:
                return error_value

        _ENC_THRESHOLD = 50

        colors['C0'] = term.normal
        colors['C1'] = term.cyan
        colors['CBold'] = term.bold
        colors['CName'] = term.blue
        colors['CTemp'] = _conditional(lambda: self.temperature < 50,
                                       term.red, term.bold_red)
        colors['FSpeed'] = _conditional(lambda: self.fan_speed < 30,
                                        term.cyan, term.bold_cyan)
        colors['CMemU'] = term.bold_yellow
        colors['CMemT'] = term.yellow
        colors['CMemP'] = term.yellow
        colors['CCPUMemU'] = term.yellow
        colors['CUser'] = term.bold_black   # gray
        colors['CUtil'] = _conditional(lambda: self.utilization < 30,
                                       term.green, term.bold_green)
        colors['CUtilEnc'] = _conditional(
            lambda: self.utilization_enc < _ENC_THRESHOLD,
            term.green, term.bold_green)
        colors['CUtilDec'] = _conditional(
            lambda: self.utilization_dec < _ENC_THRESHOLD,
            term.green, term.bold_green)
        colors['CCPUUtil'] = term.green
        colors['CPowU'] = _conditional(
            lambda: (self.power_limit is not None and
                     float(self.power_draw) / self.power_limit < 0.4),
            term.magenta, term.bold_magenta
        )
        colors['CPowL'] = term.magenta
        colors['CCmd'] = term.color(24)   # a bit dark

        if not with_colors:
            for k in list(colors.keys()):
                colors[k] = ''

        def _repr(v, none_value='??'):
            return none_value if v is None else v

        # build one-line display information
        # we want power use optional, but if deserves being grouped with
        # temperature and utilization
        reps = u"%(C1)s[{entry[index]}]%(C0)s "
        if gpuname_width is None or gpuname_width != 0:
            reps += u"%(CName)s{entry_name:{gpuname_width}}%(C0)s |"
        reps += u"%(CTemp)s{entry[temperature.gpu]:>3}°C%(C0)s, "

        if show_fan_speed:
            reps += "%(FSpeed)s{entry[fan.speed]:>3} %%%(C0)s, "

        reps += "%(CUtil)s{entry[utilization.gpu]:>3} %%%(C0)s"

        if show_codec:
            codec_info = []
            if "enc" in show_codec:
                codec_info.append(
                    "%(CBold)sE: %(C0)s"
                    "%(CUtilEnc)s{entry[utilization.enc]:>3} %%%(C0)s")
            if "dec" in show_codec:
                codec_info.append(
                    "%(CBold)sD: %(C0)s"
                    "%(CUtilDec)s{entry[utilization.dec]:>3} %%%(C0)s")
            reps += " ({})".format("  ".join(codec_info))

        if show_power:
            reps += ",  %(CPowU)s{entry[power.draw]:>3}%(C0)s "
            if show_power is True or 'limit' in show_power:
                reps += "/ %(CPowL)s{entry[enforced.power.limit]:>3}%(C0)s "
                reps += "%(CPowL)sW%(C0)s"
            else:
                reps += "%(CPowU)sW%(C0)s"

        reps += " | %(C1)s%(CMemU)s{entry[memory.used]:>5}%(C0)s " \
            "/ %(CMemT)s{entry[memory.total]:>5}%(C0)s MB"
        reps = (reps) % colors
        reps = reps.format(
            entry={k: _repr(v) for k, v in self.entry.items()},
            entry_name=util.shorten_left(
                self.entry["name"], width=gpuname_width, placeholder='…'),
            gpuname_width=gpuname_width or DEFAULT_GPUNAME_WIDTH
        )
        reps += " |"

        def process_repr(p):
            r = ''
            if not show_cmd or show_user:
                r += "{CUser}{}{C0}".format(
                    _repr(p['username'], '--'), **colors
                )
            if show_cmd:
                if r:
                    r += ':'
                r += "{C1}{}{C0}".format(
                    _repr(p.get('command', p['pid']), '--'), **colors
                )

            if show_pid:
                r += ("/%s" % _repr(p['pid'], '--'))
            r += '({CMemP}{}M{C0})'.format(
                _repr(p['gpu_memory_usage'], '?'), **colors
            )
            return r

        def full_process_info(p):
            r = "{C0} ├─ {:>6} ".format(
                    _repr(p['pid'], '--'), **colors
                )
            r += "{C0}({CCPUUtil}{:4.0f}%{C0}, {CCPUMemU}{:>6}{C0})".format(
                    _repr(p['cpu_percent'], '--'),
                    util.bytes2human(_repr(p['cpu_memory_usage'], 0)), **colors
                )
            full_command_pretty = util.prettify_commandline(
                p['full_command'], colors['C1'], colors['CCmd'])
            r += "{C0}: {CCmd}{}{C0}".format(
                _repr(full_command_pretty, '?'),
                **colors
            )
            return r

        processes = self.entry['processes']
        full_processes = []
        if processes is None:
            # None (not available)
            reps += ' ({})'.format(NOT_SUPPORTED)
        else:
            for p in processes:
                reps += ' ' + process_repr(p)
                if show_full_cmd:
                    full_processes.append(os.linesep + full_process_info(p))
        if show_full_cmd and full_processes:
            full_processes[-1] = full_processes[-1].replace('├', '└', 1)
            reps += ''.join(full_processes)
        fp.write(reps)
        return fp

    def jsonify(self):
        o = self.entry.copy()
        if self.entry['processes'] is not None:
            o['processes'] = [{k: v for (k, v) in p.items() if k != 'gpu_uuid'}
                              for p in self.entry['processes']]
        return o


class GPUStatCollection(object):

    global_processes = {}

    def __init__(self, gpu_list, driver_version=None):
        self.gpus = gpu_list

        # attach additional system information
        self.hostname = platform.node()
        self.query_time = datetime.now()
        self.driver_version = driver_version

    @staticmethod
    def clean_processes():
        for pid in list(GPUStatCollection.global_processes.keys()):
            if not psutil.pid_exists(pid):
                del GPUStatCollection.global_processes[pid]

    @staticmethod
    def new_query(debug=False):
        """Query the information of all the GPUs on local machine"""

        N.nvmlInit()
        log = util.DebugHelper()

        def _decode(b):
            if isinstance(b, bytes):
                return b.decode('utf-8')    # for python3, to unicode
            return b

        def get_gpu_info(handle):
            """Get one GPU information specified by nvml handle"""

            def get_process_info(nv_process):
                """Get the process information of specific pid"""
                process = {}
                if nv_process.pid not in GPUStatCollection.global_processes:
                    GPUStatCollection.global_processes[nv_process.pid] = \
                        psutil.Process(pid=nv_process.pid)
                ps_process = GPUStatCollection.global_processes[nv_process.pid]

                # TODO: ps_process is being cached, but the dict below is not.
                process['username'] = ps_process.username()
                # cmdline returns full path;
                # as in `ps -o comm`, get short cmdnames.
                _cmdline = ps_process.cmdline()
                if not _cmdline:
                    # sometimes, zombie or unknown (e.g. [kworker/8:2H])
                    process['command'] = '?'
                    process['full_command'] = ['?']
                else:
                    process['command'] = os.path.basename(_cmdline[0])
                    process['full_command'] = _cmdline
                # Bytes to MBytes
                # if drivers are not TTC this will be None.
                usedmem = nv_process.usedGpuMemory // MB if \
                          nv_process.usedGpuMemory else None
                process['gpu_memory_usage'] = usedmem
                process['cpu_percent'] = ps_process.cpu_percent()
                process['cpu_memory_usage'] = \
                    round((ps_process.memory_percent() / 100.0) *
                          psutil.virtual_memory().total)
                process['pid'] = nv_process.pid
                return process

            name = _decode(N.nvmlDeviceGetName(handle))
            uuid = _decode(N.nvmlDeviceGetUUID(handle))

            try:
                temperature = N.nvmlDeviceGetTemperature(
                    handle, N.NVML_TEMPERATURE_GPU
                )
            except N.NVMLError as e:
                log.add_exception("temperature", e)
                temperature = None  # Not supported

            try:
                fan_speed = N.nvmlDeviceGetFanSpeed(handle)
            except N.NVMLError as e:
                log.add_exception("fan_speed", e)
                fan_speed = None  # Not supported

            try:
                memory = N.nvmlDeviceGetMemoryInfo(handle)  # in Bytes
            except N.NVMLError as e:
                log.add_exception("memory", e)
                memory = None  # Not supported

            try:
                utilization = N.nvmlDeviceGetUtilizationRates(handle)
            except N.NVMLError as e:
                log.add_exception("utilization", e)
                utilization = None  # Not supported

            try:
                utilization_enc = N.nvmlDeviceGetEncoderUtilization(handle)
            except N.NVMLError as e:
                log.add_exception("utilization_enc", e)
                utilization_enc = None  # Not supported

            try:
                utilization_dec = N.nvmlDeviceGetDecoderUtilization(handle)
            except N.NVMLError as e:
                log.add_exception("utilization_dnc", e)
                utilization_dec = None  # Not supported

            try:
                power = N.nvmlDeviceGetPowerUsage(handle)
            except N.NVMLError as e:
                log.add_exception("power", e)
                power = None

            try:
                power_limit = N.nvmlDeviceGetEnforcedPowerLimit(handle)
            except N.NVMLError as e:
                log.add_exception("power_limit", e)
                power_limit = None

            try:
                nv_comp_processes = \
                    N.nvmlDeviceGetComputeRunningProcesses(handle)
            except N.NVMLError as e:
                log.add_exception("compute_processes", e)
                nv_comp_processes = None  # Not supported
            try:
                nv_graphics_processes = \
                    N.nvmlDeviceGetGraphicsRunningProcesses(handle)
            except N.NVMLError as e:
                log.add_exception("graphics_processes", e)
                nv_graphics_processes = None  # Not supported

            if nv_comp_processes is None and nv_graphics_processes is None:
                processes = None
            else:
                processes = []
                nv_comp_processes = nv_comp_processes or []
                nv_graphics_processes = nv_graphics_processes or []
                # A single process might run in both of graphics and compute mode,
                # However we will display the process only once
                seen_pids = set()
                for nv_process in nv_comp_processes + nv_graphics_processes:
                    if nv_process.pid in seen_pids:
                        continue
                    seen_pids.add(nv_process.pid)
                    try:
                        process = get_process_info(nv_process)
                        processes.append(process)
                    except psutil.NoSuchProcess:
                        # TODO: add some reminder for NVML broken context
                        # e.g. nvidia-smi reset  or  reboot the system
                        pass
                    except psutil.AccessDenied:
                        pass
                    except FileNotFoundError:
                        # Ignore the exception which probably has occured
                        # from psutil, due to a non-existent PID (see #95).
                        # The exception should have been translated, but
                        # there appears to be a bug of psutil. It is unlikely
                        # FileNotFoundError is thrown in different situations.
                        pass

                # TODO: Do not block if full process info is not requested
                time.sleep(0.1)
                for process in processes:
                    pid = process['pid']
                    cache_process = GPUStatCollection.global_processes[pid]
                    process['cpu_percent'] = cache_process.cpu_percent()

            index = N.nvmlDeviceGetIndex(handle)

            # GPU Info.
            # We use the same key/spec as per `nvidia-smi --query-help-gpu`
            gpu_info = {
                'index': index,
                'uuid': uuid,
                'name': name,
                'temperature.gpu': temperature,
                'fan.speed': fan_speed,
                'utilization.gpu': utilization.gpu if utilization else None,
                'utilization.enc':
                    utilization_enc[0] if utilization_enc else None,
                'utilization.dec':
                    utilization_dec[0] if utilization_dec else None,
                'power.draw': power // 1000 if power is not None else None,
                'enforced.power.limit': power_limit // 1000
                if power_limit is not None else None,
                # Convert bytes into MBytes
                'memory.used': memory.used // MB if memory else None,
                'memory.total': memory.total // MB if memory else None,
                'processes': processes,
            }
            GPUStatCollection.clean_processes()
            return gpu_info

        # 1. get the list of gpu and status
        gpu_list = []
        device_count = N.nvmlDeviceGetCount()

        for index in range(device_count):
            handle = N.nvmlDeviceGetHandleByIndex(index)
            gpu_info = get_gpu_info(handle)
            gpu_stat = GPUStat(gpu_info)
            gpu_list.append(gpu_stat)

        # 2. additional info (driver version, etc).
        try:
            driver_version = _decode(N.nvmlSystemGetDriverVersion())
        except N.NVMLError as e:
            log.add_exception("driver_version", e)
            driver_version = None    # N/A

        if debug:
            log.report_summary()

        N.nvmlShutdown()
        return GPUStatCollection(gpu_list, driver_version=driver_version)

    def __len__(self):
        return len(self.gpus)

    def __iter__(self):
        return iter(self.gpus)

    def __getitem__(self, index):
        return self.gpus[index]

    def __repr__(self):
        s = 'GPUStatCollection(host=%s, [\n' % self.hostname
        s += '\n'.join('  ' + str(g) for g in self.gpus)
        s += '\n])'
        return s

    # --- Printing Functions ---

    def print_formatted(self, fp=sys.stdout, force_color=False, no_color=False,
                        show_cmd=False, show_full_cmd=False, show_user=False,
                        show_pid=False, show_fan_speed=None,
                        show_codec="", show_power=None,
                        gpuname_width=None, show_header=True,
                        eol_char=os.linesep,
                        ):
        # ANSI color configuration
        if force_color and no_color:
            raise ValueError("--color and --no_color can't"
                             " be used at the same time")

        if force_color:
            TERM = os.getenv('TERM') or 'xterm-256color'
            t_color = Terminal(kind=TERM, force_styling=True)

            # workaround of issue #32 (watch doesn't recognize sgr0 characters)
            t_color._normal = u'\x1b[0;10m'
        elif no_color:
            t_color = Terminal(force_styling=None)
        else:
            t_color = Terminal()   # auto, depending on isatty

        # appearance settings
        if gpuname_width is None:
            gpuname_width = max([len(g.entry['name']) for g in self])

        # header
        if show_header:
            if IS_WINDOWS:
                # no localization is available; just use a reasonable default
                # same as str(timestr) but without ms
                timestr = self.query_time.strftime('%Y-%m-%d %H:%M:%S')
            else:
                time_format = locale.nl_langinfo(locale.D_T_FMT)
                timestr = self.query_time.strftime(time_format)
            header_template = '{t.bold_white}{hostname:{width}}{t.normal}  '
            header_template += '{timestr}  '
            header_template += '{t.bold_black}{driver_version}{t.normal}'

            header_msg = header_template.format(
                    hostname=self.hostname,
                    width=(gpuname_width or DEFAULT_GPUNAME_WIDTH) + 3,  # len("[?]")
                    timestr=timestr,
                    driver_version=self.driver_version,
                    t=t_color,
                )

            fp.write(header_msg.strip())
            fp.write(eol_char)

        # body
        for g in self:
            g.print_to(fp,
                       show_cmd=show_cmd,
                       show_full_cmd=show_full_cmd,
                       show_user=show_user,
                       show_pid=show_pid,
                       show_fan_speed=show_fan_speed,
                       show_codec=show_codec,
                       show_power=show_power,
                       gpuname_width=gpuname_width,
                       term=t_color)
            fp.write(eol_char)

        fp.flush()

    def jsonify(self):
        return {
            'hostname': self.hostname,
            'driver_version': self.driver_version,
            'query_time': self.query_time,
            "gpus": [g.jsonify() for g in self]
        }

    def print_json(self, fp=sys.stdout):
        def date_handler(obj):
            if hasattr(obj, 'isoformat'):
                return obj.isoformat()
            else:
                raise TypeError(type(obj))

        o = self.jsonify()
        json.dump(o, fp, indent=4, separators=(',', ': '),
                  default=date_handler)
        fp.write(os.linesep)
        fp.flush()


def new_query():
    '''
    Obtain a new GPUStatCollection instance by querying nvidia-smi
    to get the list of GPUs and running process information.
    '''
    return GPUStatCollection.new_query()



"""
The gpustat module.
"""

__version__ = '1.0.0'

from .core import GPUStat, GPUStatCollection
from .core import new_query
from .cli import print_gpustat, main


__all__ = (
    '__version__',
    'GPUStat', 'GPUStatCollection',
    'new_query',
    'print_gpustat', 'main',
)



""" Miscellaneous Utilities. """

import sys
import os.path
import traceback
import collections


def bytes2human(in_bytes):
    '''Convert bytes (int) to a human-readable string.'''
    suffixes = ('B', 'KB', 'MB', 'GB', 'TB', 'PB')
    suffix = 0
    result = int(in_bytes)
    while result > 9999 and suffix < len(suffixes):
        result = result >> 10
        suffix += 1

    if suffix >= len(suffixes):
        suffix -= 1
    return "%d%s" % (result, suffixes[suffix])


def prettify_commandline(cmdline, color_command='', color_text=''):
    '''
    Prettify and colorize a full command-line (given as list of strings),
    where command (basename) is highlighted in a different color.
    '''
    # cmdline: Iterable[str]
    if isinstance(cmdline, str):
        return cmdline
    assert cmdline

    command_p, command_b = os.path.split(cmdline[0])
    s = color_text + os.path.join(command_p,
                                  color_command + command_b + color_text)

    if len(cmdline) > 1:
        s += ' '
        s += ' '.join(cmdline[1:])
    return s


def shorten_left(text, width, placeholder="…"):
    # text: str
    if width is None:
        return text
    if text is None or len(text) <= width:
        return text
    if width < 0:
        raise ValueError("width must be non-negative.")
    if width == 0:
        return ""

    if width == len(placeholder):
        return placeholder
    elif width - len(placeholder) < 0:
        return placeholder[:width]
        # raise ValueError("width is smaller than the length of placeholder.")
    return placeholder + text[-(width - len(placeholder)):]


class DebugHelper:

    def __init__(self):
        self._reports = []

    def add_exception(self, column, e=None):
        msg = "> An error while retrieving `{column}`: {e}".format(
            column=column, e=str(e))
        self._reports.append((msg, e))

    def _write(self, msg):
        sys.stderr.write(msg)
        sys.stderr.write('\n')

    def report_summary(self, concise=True):
        _seen_messages = collections.defaultdict(int)
        for msg, e in self._reports:
            if msg not in _seen_messages or not concise:
                self._write(msg)
                self._write(''.join(traceback.format_exception(None, e, e.__traceback__)))
            _seen_messages[msg] += 1

        if concise:
            for msg, value in _seen_messages.items():
                self._write("{msg} -> Total {value} occurrences.".format(
                    msg=msg, value=value))
            self._write('')



"""Imports pynvml with sanity checks and custom patches."""

import textwrap
import os


pynvml = None

# If this environment variable is set, we will bypass pynvml version validation
# so that legacy pynvml (nvidia-ml-py3) can be used. This would be useful
# in a case where there are conflicts on pynvml dependencies.
# However, beware that pynvml might produce wrong results (see #107).
ALLOW_LEGACY_PYNVML = os.getenv("ALLOW_LEGACY_PYNVML", "")
ALLOW_LEGACY_PYNVML = ALLOW_LEGACY_PYNVML.lower() not in ('false', '0', '')


try:
    # Check pynvml version: we require 11.450.129 or newer.
    # https://github.com/wookayin/gpustat/pull/107
    import pynvml
    if not (
        # Requires nvidia-ml-py >= 11.460.79
        hasattr(pynvml, 'NVML_BRAND_NVIDIA_RTX') or
        # Requires nvidia-ml-py >= 11.450.129, < 11.510.69
        hasattr(pynvml, 'nvmlDeviceGetComputeRunningProcesses_v2')
    ) and not ALLOW_LEGACY_PYNVML:
        raise RuntimeError("pynvml library is outdated.")
except (ImportError, SyntaxError, RuntimeError) as e:
    raise ImportError(textwrap.dedent(
        """\
        pynvml is missing or an outdated version is installed.

        We require nvidia-ml-py>=11.450.129, and nvidia-ml-py3 shall not be used.
        For more details, please refer to: https://github.com/wookayin/gpustat/issues/107

        Your pynvml installation: """ + repr(pynvml) +
        """

        -----------------------------------------------------------
        Please reinstall `gpustat`:

        $ pip install --force-reinstall gpustat

        If it still does not fix the problem, please manually fix nvidia-ml-py installation:

        $ pip uninstall nvidia-ml-py3
        $ pip install --force-reinstall 'nvidia-ml-py<=11.495.46'
        """)) from e


__all__ = ['pynvml']
